from enum import Enum

import logging


log = logging.getLogger(__name__)


def on_event(*events):
    """
    Nyuki method decorator to register a callback for a bus event.
    """
    def call(func):
        func.on_event = set(events)
        return func
    return call


class Event(Enum):
    """
    Bus events that can be trigger upon reception of an XMPP event.
    """
    Connecting = 'connecting'
    Connected = 'connected'
    ConnectionError = 'connection_error'
    Disconnected = 'disconnected'
    EventReceived = 'event_received'


class EventManager(object):
    """
    Manager that helps to perform asynchronous callbacks when a bus event is
    triggered. This class use our asyncio event loop wrapper EventLoop.
    """
    def __init__(self, loop):
        self._loop = loop
        self._callbacks = self._init_callbacks()
        log.debug("Available events: {}".format(list(Event)))
        log.debug("Events will be called through {}".format(self._loop))

    @staticmethod
    def _init_callbacks():
        """
        Build the main data structure with available events.
        """
        callbacks = dict()
        for event in Event:
            callbacks[event] = set()
        return callbacks

    def register(self, event, callback):
        """
        Add a callback for a given event.
        """
        if event not in Event:
            raise ValueError("Event {} does not exist".format(event))
        self._callbacks[event].add(callback)
        log.debug("Callback added for {} event".format(event.name))

    def trigger(self, event, *args):
        """
        Perform all registered callbacks for the triggered event.
        """
        if event not in self._callbacks:
            raise ValueError("Event {} is not registered".format(event))
        for callback in self._callbacks[event]:
            self._loop.call_soon(callback, *args)
        log.debug("{} event triggered".format(event.name))



import json
import logging
import asyncio

from nyuki.api import Api


log = logging.getLogger(__name__)


def resource(endpoint, version=None):

    """
    Nyuki resource decorator to register a route.
    A resource has multiple HTTP methods (get, post, etc).
    """

    def decorated(cls):
        cls.endpoint = endpoint
        cls.version = version
        return cls
    return decorated


class Capability(object):
    """
    A capability is unique (hashable object, based on capability's name).
    """
    def __init__(self, name, method, endpoint, version, handler, wrapper):
        self.name = name
        self.method = method
        self.endpoint = endpoint
        self.version = version
        self.handler = handler
        self.wrapper = wrapper

    def __hash__(self):
        return hash(self.name)


class Response(object):

    """
    This class is a generic response (with a body and a status) that can be
    used by either the bus or the API.
    """

    ENCODING = 'utf-8'

    def __init__(self, body=None, status=200):
        self.body = body
        self.status = status
        self.headers = {}
        self._validate()

    def _validate(self):
        self._validate_body()
        self._validate_status()

    def _validate_body(self):
        if not self.body:
            return

        if not isinstance(self.body, dict) and not isinstance(self.body, list):
            raise ValueError("Response body should be a dictionary or a list")

        # If body and body is json, set the right headers
        self.headers = {'Content-Type': 'application/json'}

    def _validate_status(self):
        if not isinstance(self.status, int):
            raise ValueError("Response status code should be a integer")

    @property
    def api_payload(self):
        """
        Used by the HTTP API.
        """
        if self.body:
            payload = json.dumps(self.body)
            return bytes(payload, self.ENCODING)
        return None


class Exposer(object):

    """
    Provide a engine to expose nyuki capabilities through a HTTP API.
    """

    def __init__(self, loop, debug=False):
        self._loop = loop
        self._capabilities = set()
        self._api = Api(loop, debug=debug)
        log.debug("Capabilities will be called through {}".format(loop))

    @property
    def capabilities(self):
        return self._capabilities

    def register(self, capa):
        """
        Add a capability and register its route to the API.
        """
        if capa in self._capabilities:
            raise ValueError("A capability is already exposed through {} with "
                             "{} method".format(capa.endpoint, capa.method))
        self._capabilities.add(capa)
        # Handle version as part of the endpoint (if defined)
        if capa.version:
            endpoint = '/{}{}'.format(capa.version, capa.endpoint)
        else:
            endpoint = capa.endpoint
        self._api.router.add_route(capa.method, endpoint, capa.wrapper)
        log.debug("Capability added: {}".format(capa.name))

    def expose(self, host='0.0.0.0', port=8080):
        """
        Expose capabilities by building the HTTP server.
        The server will be started with the event loop.
        """
        self._loop.run_until_complete(self._api.build(host, port))

    def restart(self, host='0.0.0.0', port=8080):
        """
        Restart the HTTP server.
        """
        # Stopping the API server
        fut = self.shutdown()
        # Rebuilding the API server afterwards
        fut.add_done_callback(
            lambda x: asyncio.async(self._api.build(host, port)))

    def shutdown(self):
        """
        Shutdown capabilities exposure by destroying the HTTP server.
        """
        log.debug("Stopping the http server")
        return asyncio.async(self._api.destroy(), loop=self._loop)

    def _find(self, capa_name):
        """
        Get a capability by its name.
        """
        for capability in self._capabilities:
            if capa_name == capability.name:
                return capability

    def call(self, name, request):
        """
        Call a capability by its name in an asynchronous fashion.
        """
        capa = self._find(name)
        if not capa:
            log.warning("Capability {} is called but doen't exist".format(name))
            return

        # TODO: use asyncio.ensure_future when Python 3.4.4 will be released
        future = asyncio.async(capa.wrapper(request), loop=self._loop)
        return future



DEFAULT_LOGGING = {
    "version": 1,
    "formatters": {
        "long": {
            "format": "%(asctime)-24s %(levelname)-8s [%(name)s] %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "long",
            "stream": "ext://sys.stdout"
        },
        "loop_stderr": {
            "class": "logging.StreamHandler",
            "formatter": "long",
            "stream": "ext://sys.stderr",
            "level": "ERROR"
        }
    },
    "root": {
        "handlers": ["console"],
        "level": "INFO"
    },
    "loggers": {
        "asyncio": {
            "handlers": ["loop_stderr"],
            "level": "WARNING"
        },
        "slixmpp": {
            "level": "WARNING"
        },
        "aiohttp.web": {
            "level": "WARNING"
        }
    },
    "disable_existing_loggers": False
}



import logging
import asyncio
from inspect import getmembers, isclass, isfunction

from nyuki.api import Method
from nyuki.capabilities import Capability


log = logging.getLogger(__name__)


class CapabilityHandler(type):
    ALLOWED_METHODS = Method.list()

    def __call__(cls, *args, **kwargs):
        """
        Register decorated resources and methods to be routed by the web app.
        """
        nyuki = super().__call__(*args, **kwargs)
        for resrc, desc in cls._filter_resource(nyuki):
            version = desc.version
            endpoint = desc.endpoint
            for method, handler in cls._filter_capability(desc):
                name = '{}_{}'.format(method, resrc)
                wrapper = cls._build_wrapper(nyuki, handler)
                nyuki.capability_exposer.register(Capability(
                    name=name.lower(),
                    method=method,
                    endpoint=endpoint,
                    version=version,
                    handler=handler,
                    wrapper=wrapper
                ))
        return nyuki

    @staticmethod
    def _build_wrapper(obj, func):
        """
        Build a wrapper method to be called by the web server.
        Route callbacks are supposed to be called through `func(request)`,
        the following code updates capabilities to be executed as instance
        methods: `func(nyuki, request)`.
        """
        return asyncio.coroutine(
            lambda req, **kwargs: func(obj, req, **kwargs)
        )

    @classmethod
    def _filter_capability(mcs, resrc):
        """
        Find methods decorated with `capability`.
        """
        for name, handler in getmembers(resrc, isfunction):
            method = name.upper()
            if method not in mcs.ALLOWED_METHODS:
                raise ValueError("{} is not a valid HTTP method".format(method))
            yield method, handler

    @staticmethod
    def _filter_resource(obj):
        """
        Find nested classes decorated with `endpoint`.
        """
        for name, cls in getmembers(obj, isclass):
            if hasattr(cls, 'endpoint'):
                yield name, cls


class EventHandler(type):
    def __call__(cls, *args, **kwargs):
        """
        Register decorated method to be called when an event is trigger.
        """
        nyuki = super().__call__(*args, **kwargs)
        for method, events in cls._filter_event(nyuki):
            for event in events:
                nyuki.event_manager.register(event, method)
        return nyuki

    @staticmethod
    def _filter_event(obj):
        """
        Find methods decorated with `on_event`.
        """
        for attr in dir(obj):
            value = getattr(obj, attr)
            if callable(value) and hasattr(value, 'on_event'):
                yield value, value.on_event


class MetaHandler(EventHandler, CapabilityHandler):
    """
    Meta class that registers all decorated methods as either a capability or
    a callback for a bus event.
    """
    def __call__(cls, *args, **kwargs):
        nyuki = super().__call__(*args, **kwargs)
        return nyuki



import asyncio
import codecs
import json
from jsonschema import validate, ValidationError
import logging
import logging.config
import signal
import sys

from nyuki.bus import Bus
from nyuki.capabilities import Exposer, Response, resource
from nyuki.commands import get_command_kwargs
from nyuki.config import (
    get_full_config, write_conf_json, merge_configs, DEFAULT_CONF_FILE
)
from nyuki.events import EventManager
from nyuki.handlers import MetaHandler


log = logging.getLogger(__name__)


class Nyuki(metaclass=MetaHandler):

    """
    A lightweigh base class to build nyukis. A nyuki provides tools that shall
    help the developer with managing the following topics:
      - Bus of communication between nyukis.
      - Asynchronous events.
      - Capabilities exposure through a REST API.
    This class has been written to perform the features above in a reliable,
    single-threaded, asynchronous and concurrent-safe environment.
    The core engine of a nyuki implementation is the asyncio event loop
    (a single loop is used for all features).
    A wrapper is also provide to ease the use of asynchronous calls
    over the actions nyukis are inteded to do.
    """

    # Configuration schema must follow jsonschema rules.
    BASE_CONF_SCHEMA = {
        "type": "object",
        "required": ["bus", "api", "log"],
        "properties": {
            "bus": {
                "type": "object",
                "required": ["jid", "password"],
                "properties": {
                    "jid": {"type": "string"},
                    "password": {"type": "string"}
                }
            }
        }
    }

    def __init__(self, **kwargs):
        # Set stdout as utf-8 codec
        try:
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout.detach())
        except Exception as exc:
            # Nosetests seems to alter stdout, breaking detach()
            log.warning('Could not change stdout codec')
            log.exception(exc)

        # List of configuration schemas
        self._schemas = []

        # Get configuration from multiple sources and register base schema
        kwargs = kwargs or get_command_kwargs()
        self.config_filename = kwargs.get('config', DEFAULT_CONF_FILE)
        self._config = get_full_config(**kwargs)
        self.register_schema(self.BASE_CONF_SCHEMA)

        # Initialize logging
        logging.config.dictConfig(self._config['log'])

        self.loop = asyncio.get_event_loop()
        self.event_manager = EventManager(self.loop)
        self._bus = self._make_bus()
        self._exposer = Exposer(
            self.loop,
            debug=(self._config['log']['root']['level'] == "DEBUG")
        )

        self.is_stopping = False

    @property
    def config(self):
        return self._config

    @property
    def capabilities(self):
        return self._exposer.capabilities

    @property
    def capability_exposer(self):
        return self._exposer

    @property
    def request(self):
        return self._bus.request

    @property
    def publish(self):
        return self._bus.publish

    @property
    def subscribe(self):
        return self._bus.subscribe

    def _make_bus(self):
        """
        Returns a new set up Bus object.
        """
        bus = Bus(
            loop=self.loop,
            event_manager=self.event_manager,
            **self._config['bus']
        )
        bus.client.disconnected.add_done_callback(self._bus_disconnected)
        return bus

    def start(self):
        """
        Start the nyuki: launch the bus client and expose capabilities.
        Basically, it starts the event loop.
        """
        signal.signal(signal.SIGTERM, self.abort)
        signal.signal(signal.SIGINT, self.abort)
        self._bus.connect()
        self._exposer.expose(**self._config['api'])

        if asyncio.iscoroutinefunction(self.setup):
            asyncio.async(self.setup())
        else:
            self.loop.call_soon(self.setup)

        self.loop.run_forever()
        self.loop.close()

    def _bus_disconnected(self, future):
        """
        Handle bus disconnection.
        If it was manual stop the nyuki, else try to reconnect.
        """
        log.info('Bus disconnected')
        if self.is_stopping:
            log.info('Tearing down nyuki')
            self.teardown()
            self._stop_loop()
        else:
            log.info('Reconnecting...')
            self._bus = self._make_bus()
            self._bus.connect()

    def _stop_loop(self):
        """
        Call the loop to stop itself.
        """
        self.loop.call_soon_threadsafe(self.loop.stop)

    def abort(self, signum, frame):
        """
        Signal handler: gracefully stop the nyuki.
        """
        log.warning("Caught signal {}".format(signum))
        self.stop()

    def stop(self, timeout=5):
        """
        Stop the nyuki. Basically, disconnect to the bus. That will eventually
        trigger a `Disconnected` event.
        """
        if self.is_stopping:
            log.warning('Forcing the nyuki stopping')
            self._stop_loop()
            return

        self.is_stopping = True

        def timed_out():
            log.warning('Could not stop nyuki after '
                        '{} seconds, killing'.format(timeout))
            self._stop_loop()

        self.loop.call_later(timeout, timed_out)
        self._exposer.shutdown()
        self._bus.disconnect()

    def register_schema(self, schema, format_checker=None):
        """
        Add a jsonschema to validate on configuration update.
        """
        self._schemas.append((schema, format_checker))
        self._validate_config()

    def _validate_config(self, config=None):
        """
        Validate on all registered configuration schemas.
        """
        log.debug('Validating configuration')
        config = config or self._config
        for schema, checker in self._schemas:
            validate(config, schema, format_checker=checker)

    @asyncio.coroutine
    def setup(self):
        """
        First thing called when starting the event loop, coroutine or not.
        """
        pass

    def teardown(self):
        """
        Called right before closing the event loop, stopping the Nyuki.
        """
        pass

    def update_config(self, *new_confs):
        """
        Update the current configuration with the given list of dicts.
        """
        config = merge_configs(self._config, *new_confs)
        self._validate_config(config)
        self._config = config

    def save_config(self):
        """
        Save the current configuration dict to its JSON file.
        """
        write_conf_json(self.config, self.config_filename)

    def reload(self, services=False):
        """
        Override this to implement a reloading to your Nyuki.
        (called on PATCH /config)
        """
        self.save_config()
        logging.config.dictConfig(self._config['log'])
        if services:
            self._bus.disconnect()
            self._exposer.restart(**self._config['api'])

    @resource(endpoint='/config', version='v1')
    class Configuration:

        def get(self, request):
            return Response(self._config)

        def patch(self, request):
            try:
                self.update_config(request)
            except ValidationError as error:
                error = {'error': error.message}
                log.error('Bad configuration received : {}'.format(request))
                log.debug(error)
                return Response(body=error, status=400)
            else:
                self.reload('api' in request or 'bus' in request)
            return Response(self._config)

    @resource(endpoint='/swagger', version='v1')
    class Swagger:
        def get(self, request):
            try:
                with open('swagger.json', 'r') as f:
                    body = json.loads(f.read())
            except OSError:
                return Response(status=404, body={
                    'error': 'Missing swagger documentation'
                })

            return Response(body=body)



import aiohttp
import asyncio
import json
import logging
from slixmpp import ClientXMPP
from slixmpp.exceptions import IqError, IqTimeout

from nyuki.events import Event


log = logging.getLogger(__name__)


class _BusClient(ClientXMPP):
    """
    XMPP client to connect to the bus.
    This class is based on Slixmpp (fork of Sleexmpp) using asyncio event loop.
    """
    def __init__(self, jid, password, host=None, port=5222):
        super().__init__(jid, password)
        host = host or self.boundjid.domain
        self._address = (host, port)

        self.register_plugin('xep_0045')  # Multi-user chat
        self.register_plugin('xep_0077')  # In-band registration

        # Disable IPv6 until we really need it
        self.use_ipv6 = False

    def connect(self, **kwargs):
        """
        Connect to the XMPP server using the default address computed at init
        if not passed as argument.
        """
        try:
            self._address = kwargs.pop('address')
        except KeyError:
            pass
        super().connect(address=self._address, **kwargs)
        self.init_plugins()


class Bus(object):
    """
    A simple class that implements Publish/Subscribe-like communications over
    XMPP. This communication layer is also called "bus".
    Each nyuki has its own "topic" (actually a MUC) to publish events. And it
    can publish events to that topic only. In the meantime, each nyuki can
    subscribe to any other topics to process events from other nyukis.
    """

    MUC_DOMAIN = 'mucs.localhost'

    def __init__(self, jid, password, host=None, port=5222, loop=None,
                 event_manager=None):
        self._loop = loop
        self.client = _BusClient(jid, password, host, port)
        self.client.loop = self._loop
        self.client.add_event_handler('connection_failed', self._on_failure)
        self.client.add_event_handler('disconnected', self._on_disconnect)
        self.client.add_event_handler('groupchat_invite', self._on_invite)
        self.client.add_event_handler('groupchat_message', self._on_event)
        self.client.add_event_handler('register', self._on_register)
        self.client.add_event_handler('session_start', self._on_start)

        self.event_manager = event_manager

        self._topic = self.client.boundjid.user
        self._mucs = self.client.plugin['xep_0045']

    def _muc_address(self, topic):
        return '{}@{}'.format(topic, self.MUC_DOMAIN)

    def _on_register(self, event):
        """
        XMPP event handler while registering a user (in-band registration).
        Does not trigger bus event (internal behavior), except if it fails.
        """
        def done(future):
            try:
                future.result()
            except IqError as exc:
                error = exc.iq['error']['text']
                log.debug("Could not register account: {}".format(error))
            except IqTimeout:
                log.error("No response from the server")
                self.event_manager.trigger(Event.ConnectionError)
            finally:
                log.debug("Account {} created".format(self.client.boundjid))

        resp = self.client.Iq()
        resp['type'] = 'set'
        resp['register']['username'] = self.client.boundjid.user
        resp['register']['password'] = self.client.password
        future = resp.send()
        future.add_done_callback(done)

    def _on_start(self, event):
        """
        XMPP event handler when the connection has been made.
        Also trigger a bus event: `Connected`.
        """
        log.info('Connected to XMPP server {}:{}'.format(
                 *self.client._address))
        self.client.send_presence()
        self.client.get_roster()
        # Auto-subscribe to the topic where the nyuki could publish events.
        self.subscribe(self._topic)
        self.event_manager.trigger(Event.Connected)

    def _on_disconnect(self, event):
        """
        XMPP event handler when the client has been disconnected.
        Also trigger a bus event: `Disconnected`.
        """
        log.debug("Disconnected from the bus")
        self.event_manager.trigger(Event.Disconnected)

    def _on_failure(self, event):
        """
        XMPP event handler when something is going wrong with the connection.
        Also trigger a bus event: `ConnectionError`.
        """
        log.error("Connection to XMPP server {}:{} failed".format(
                  *self.client._address))
        self.client.abort()
        self.event_manager.trigger(Event.ConnectionError)

    def _on_invite(self, event):
        """
        Enter room on invitation.
        """
        self.subscribe(event['from'])

    def _on_event(self, event):
        """
        XMPP event handler when a Nyuki Event has been received.
        Fire EventReceived with (from, body).
        """
        # ignore events from the nyuki itself
        if event['from'].user == self.client.boundjid.user:
            return
        log.debug('event received: {}'.format(event))
        try:
            body = json.loads(event['body'])
        except ValueError:
            body = {}
        self.event_manager.trigger(
            Event.EventReceived, event['from'], body)

    def connect(self):
        """
        Connect to the bus.
        """
        self.client.connect()

    def disconnect(self, wait=2):
        """
        Disconnect from the bus with a default timeout set to 5s.
        """
        if not self.client.transport:
            log.warning('XMPP client is already disconnected')
            self.client.disconnected.set_result(True)
        else:
            self.client.disconnect(wait=wait)

    def publish(self, event):
        """
        Send an event in the nyuki's own MUC so that other nyukis that joined
        the MUC can process it.
        """
        if not isinstance(event, dict):
            raise TypeError('Message must be a dict')
        log.debug(">> publishing to '{}': {}".format(self._topic, event))
        msg = self.client.Message()
        msg['type'] = 'groupchat'
        msg['to'] = self._muc_address(self._topic)
        msg['body'] = json.dumps(event)
        msg.send()

    def subscribe(self, topic):
        """
        Enter into a new chatroom using xep_0045.
        Room address format must be like '{topic}@mucs.localhost'
        """
        self._mucs.joinMUC(self._muc_address(topic), self._topic)
        log.info("subscribed to '{}'".format(topic))

    def unsubscribe(self, topic):
        """
        Leave a chatroom.
        """
        self._mucs.leaveMUC(self._muc_address(topic), self._topic)
        log.info("unsubscribed to '{}'".format(topic))

    @asyncio.coroutine
    def _execute_request(self, url, method, data=None, headers=None):
        """
        Asynchronously send a request of type 'application/json' to method/url.
        If data is not None, it is supposed to be a dict.
        """
        data = json.dumps(data)
        base_headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        headers = headers or {}
        headers.update(base_headers)
        log.debug('>> sending {} request to {}'.format(method.upper(), url))
        response = yield from aiohttp.request(method, url, data=data,
                                              headers=headers)
        try:
            data = yield from response.json()
        except ValueError:
            data = None
        response.json = data
        log.debug('<< received response from {}: {}'.format(url, data))
        return response

    @asyncio.coroutine
    def request(self, nyuki, endpoint, method, out=False,
                data=None, headers=None, callback=None):
        """
        Send a P2P request to another nyuki, async a callback if given.
        The callback is called with the response object (json).
        """
        if not out:
            endpoint = 'http://localhost:8080/{}/api/{}'.format(nyuki,
                                                                endpoint)
        try:
            response = yield from self._execute_request(endpoint, method,
                                                        data, headers)
        except (aiohttp.HttpProcessingError,
                aiohttp.ServerDisconnectedError,
                aiohttp.ClientOSError) as exc:
            log.error('failed to send request to {}'.format(endpoint))
            response = exc
            error = {'error': repr(exc), 'endpoint': endpoint, 'data': data}
            self.publish(error)
        if callback:
            log.debug('calling response callback with {}'.format(response))
            if asyncio.iscoroutinefunction(callback):
                asyncio.async(callback(response))
            else:
                self._loop.call_soon(callback, response)
        return response



import logging
from argparse import ArgumentParser


log = logging.getLogger(__name__)


def _build_args():
    """
    Build argument parser and actually parse them at runtime.
    """
    parser = ArgumentParser(description='Nyuki implementation')
    parser.add_argument('-l', '--logging',
                        help='debug mode', required=False, default='INFO',
                        choices=[
                            'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
                        ])
    parser.add_argument('-c', '--config',
                        help='config file name', required=False)
    parser.add_argument('-j', '--jid',
                        help='xmpp jid: <user>[@<host>]', required=False)
    parser.add_argument('-p', '--password',
                        help='xmpp password', required=False)
    parser.add_argument('-s', '--server',
                        help='xmpp server: <host>[:<port>]', required=False)
    parser.add_argument('-a', '--api',
                        help='api binding: <host>[:<port>]', required=False)
    return parser.parse_args()


def get_command_kwargs():
    args = _build_args()

    command_args = {
        'bus': {},
        'api': {}
    }

    if args.jid:
        command_args['bus'].update(jid=args.jid)
    if args.password:
        command_args['bus'].update(password=args.password)

    # Split XMPP host/port
    if args.server:
        server = args.server.split(':')
        try:
            command_args['bus'].update(host=server[0], port=int(server[1]))
        except IndexError:
            command_args['bus'].update(host=args.server)

    # Split API host/port
    if args.api:
        api = args.api.split(':')
        try:
            command_args['api'].update(host=api[0], port=int(api[1]))
        except IndexError:
            command_args['api'].update(host=args.api)

    # Set logging root level
    command_args['log'] = {
        'root': {
            'level': args.logging
        }
    }

    if args.config:
        command_args['config'] = args.config

    return command_args



from aiohttp import web, HttpBadRequest
import asyncio
from enum import Enum
import logging


log = logging.getLogger(__name__)
# aiohttp needs its own logger, and always prints HTTP hits using INFO level
access_log = logging.getLogger('.'.join([__name__, 'access']))
access_log.info = access_log.debug


class Method(Enum):
    """
    Supported HTTP methods by the REST API.
    """
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    CONNECT = 'CONNECT'
    PATCH = 'PATCH'

    @classmethod
    def list(cls):
        return [method.name for method in cls]


class Api(object):
    """
    The goal of this class is to gather all http-related behaviours.
    Basically, it's a wrapper around aiohttp.
    """
    def __init__(self, loop, debug=False, **kwargs):
        self._loop = loop
        self._server = None
        self._handler = None
        self._middlewares = [mw_json, mw_capability]  # Call order = list order
        self._debug = debug
        self._app = web.Application(
            loop=self._loop,
            middlewares=self._middlewares,
            **kwargs
        )

    @property
    def router(self):
        return self._app.router

    @asyncio.coroutine
    def build(self, host, port):
        """
        Create a HTTP server to expose the API.
        """
        log.info("Starting the http server on {}:{}".format(host, port))
        self._handler = self._app.make_handler(
            log=log, access_log=access_log, debug=self._debug
        )
        self._server = yield from self._loop.create_server(
            self._handler, host=host, port=port
        )

    @asyncio.coroutine
    def destroy(self):
        """
        Gracefully destroy the HTTP server by closing all pending connections.
        """
        self._server.close()
        yield from self._handler.finish_connections()
        yield from self._server.wait_closed()
        log.info('Stopped the http server')


@asyncio.coroutine
def mw_json(app, next_handler):
    """
    Ensure the content type is `application/json` for all POST-like requests.
    """
    @asyncio.coroutine
    def middleware(request):
        if request.method in request.POST_METHODS:
            content_type = request.headers.get('CONTENT-TYPE')
            if content_type:
                if 'application/json' in content_type:
                    # Checking that the content is actually JSON
                    # there could be a charset specified.
                    try:
                        yield from request.json()
                    except ValueError:
                        log.error('Invalid JSON request content')
                        raise HttpBadRequest('Invalid JSON request content')
                else:
                    log.error('Invalid Content-Type')
                    raise HttpBadRequest('Invalid Content-Type')
            elif (yield from request.content.read()):
                log.error('Missing suitable Content-Type')
                raise HttpBadRequest('Missing suitable Content-Type')
        response = yield from next_handler(request)
        return response
    return middleware


class APIRequest(dict):
    """
    Class that stores the request data and the headers as an attribute.
    """

    headers = None

    @classmethod
    @asyncio.coroutine
    def from_request(cls, request):
        # Get json payload if there is one
        if request.method in request.POST_METHODS:
            try:
                data = yield from request.json()
            except ValueError:
                data = None
        else:
            data = dict(getattr(request, request.method))

        # Set up class and headers as request attribute
        req = cls(**data) if data else cls()
        req.headers = request.headers
        return req


@asyncio.coroutine
def mw_capability(app, capa_handler):
    """
    Transform the request data to be passed through a capability and
    convert the result into a web response.
    From here, we are expecting a JSON content.
    """
    @asyncio.coroutine
    def middleware(request):
        api_req = yield from APIRequest.from_request(request)
        capa_resp = yield from capa_handler(api_req, **request.match_info)

        return web.Response(
            body=capa_resp.api_payload if capa_resp else None,
            status=capa_resp.status if capa_resp else 200,
            headers=capa_resp.headers if capa_resp else {},
        )

    return middleware



#!/usr/bin/env python

from pip.req import parse_requirements
from setuptools import setup, find_packages


with open('VERSION.txt', 'r') as v:
    version = v.read().strip()

with open('DESCRIPTION', 'r') as d:
    long_description = d.read()

# Requirements
install_reqs = parse_requirements('requirements.txt', session='dummy')
reqs = [str(ir.req) for ir in install_reqs]

setup(
    name='nyuki',
    description='Allowing the creation of independent unit to deal with stream processing while exposing an XMPP and REST API.',
    long_description=long_description,
    url='http://www.surycat.com',
    author='Optiflows R&D',
    author_email='rand@surycat.com',
    version=version,
    install_requires=reqs,
    packages=find_packages(exclude=['tests']),
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
    ],
)
